/*
	ColourWarp plugin for Avisynth

	Copyright (c) 2019 David Horman, All Rights Reserved

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "windows.h"
#include "avisynth.h"
#include "Matrix.h"
#include <vector>

void __debug(const char* a, ...) {
	char string[4096];
	va_list args;

	va_start(args, a);
	vsprintf_s(string, a, args);
	va_end(args);

	OutputDebugString(string);
}

class ColourWarp : public GenericVideoFilter {
private:
	struct point_pair {
		int u, v;
		int ou, ov;
	};
	struct byte_pair {
		byte u, v;
	};
	byte_pair* uv_map;
	bool debug;
	PVideoFrame debug_frame;
	std::vector<point_pair> points;
	__inline void Warp(int x, int y, double& ox, double& oy);
	Matrix* vox = NULL;
	Matrix* voy = NULL;
	Matrix* L = NULL;
	Matrix* vx = NULL;
	Matrix* vy = NULL;

public:
	ColourWarp(PClip _child, AVSValue uv_array, AVSValue clip_array, AVSValue pixel_array, bool invert, bool _debug, IScriptEnvironment* env);
	~ColourWarp();
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};

__inline void ColourWarp::Warp(int u, int v, double& ou, double& ov) {
	ou = vox->table[(int)points.size()][0] + vox->table[(int)points.size() + 1][0] * u + vox->table[(int)points.size() + 2][0] * v;
	ov = voy->table[(int)points.size()][0] + voy->table[(int)points.size() + 1][0] * u + voy->table[(int)points.size() + 2][0] * v;

	double t, d;

	for (auto pt = points.begin(); pt < points.end(); ++pt) {// i = 0; i < (int)points.size(); i++) {
		t = pt->u - u;
		d = t * t;
		t = pt->v - v;
		d += t * t;
		if (d > 0) d = d * log(d);

		int i = (int)(pt - points.begin());
//		if (d > 0) d = d * log(d)*0.5f; // 0.5f because we want d^2log(d), not d^2(log(d^2)) - maybe it doesn't make any difference
		ou += vox->table[i][0] * d;
		ov += voy->table[i][0] * d;
	}

	ou += u;
	ov += v;
}

ColourWarp::ColourWarp(PClip _child, AVSValue uv_array, AVSValue clip_array, AVSValue pixel_array, bool invert, bool _debug, IScriptEnvironment* env) : GenericVideoFilter(_child), debug(_debug) {
	for (int i = 0; i < clip_array.ArraySize() + 1; ++i) {
		VideoInfo temp = i == 0 ? vi : clip_array[i - 1].AsClip()->GetVideoInfo();
		if (!temp.IsPlanar()) env->ThrowError("ColourWarp: Planar video only");
		if (!temp.IsYUV()) env->ThrowError("ColourWarp: YUV video only");
		if (temp.BitsPerComponent() != 8) env->ThrowError("ColourWarp: Only 8-bit video is supported");
	}

	int n_points;

	if (uv_array.ArraySize() & 3) env->ThrowError("ColourWarp: UV array size must be a multiple of 4");
	n_points = uv_array.ArraySize() >> 2;
	for (int i = 0; i < n_points; ++i) {
		int o = i << 2;
		points.push_back({
			uv_array[o].AsInt() + 0x80,
			uv_array[o + 1].AsInt() + 0x80,
			uv_array[o + 2].AsInt() - uv_array[o + 0].AsInt(),
			uv_array[o + 3].AsInt() - uv_array[o + 1].AsInt()
		});
	}

	if (pixel_array.ArraySize()) {
		if (pixel_array.ArraySize() % 6) env->ThrowError("ColourWarp: Pixel array size must be a multiple of 6");
		PClip clips[2] = { clip_array.ArraySize() == 1 ? child : clip_array[0].AsClip(), clip_array[clip_array.ArraySize() - 1].AsClip() };
		int sub_w[2];
		int sub_h[2];
		int width[2];
		int height[2];

		for (int i = 0; i < 2; ++i) {
			sub_w[i] = clips[i]->GetVideoInfo().GetPlaneWidthSubsampling(PLANAR_U);
			sub_h[i] = clips[i]->GetVideoInfo().GetPlaneHeightSubsampling(PLANAR_U);
			width[i] = clips[i]->GetVideoInfo().width >> sub_w[i];
			height[i] = clips[i]->GetVideoInfo().height >> sub_h[i];
		}

		n_points = pixel_array.ArraySize() / 3;
		int ou, ov;
		for (int i = 0; i < n_points; ++i) {
			int o = i * 3;
			PVideoFrame frame = clips[i & 1]->GetFrame(pixel_array[o].AsInt(), env);
			int pitch_u = frame->GetPitch(PLANAR_U);
			int pitch_v = frame->GetPitch(PLANAR_V);
			int x = pixel_array[o + 1].AsInt() >> sub_w[i & 1];
			int y = pixel_array[o + 2].AsInt() >> sub_h[i & 1];
			if (x < 0) x = 0; else if (x >= width[i & 1]) x = width[i & 1] - 1;
			if (y < 0) y = 0; else if (y >= height[i & 1]) y = height[i & 1] - 1;
			int u = ((const byte*)frame->GetReadPtr(PLANAR_U))[y*pitch_u + x];
			int v = ((const byte*)frame->GetReadPtr(PLANAR_V))[y*pitch_v + x];
			if (!(i & 1)) {
				ou = u;
				ov = v;
			} else {
				points.push_back({
					ou,
					ov,
					u - ou,
					v - ov
				});
			}
		}
	}

	switch (points.size()) {
		case 0: env->ThrowError("ColourWarp: No colour points defined"); break;
		case 1: {
			points.push_back({
				points[0].u + 128,
				points[0].v,
				points[0].ou,
				points[0].ov
			});
		} // no break; fall through
		case 2: {
			points.push_back({
				points[1].u + (points[1].v - points[0].v),
				points[1].v - (points[1].u - points[0].u),
				(points[1].u + points[1].ou) + ((points[1].v + points[1].ov) - (points[0].v + points[0].ov)) - (points[1].u + (points[1].v - points[0].v)),
				(points[1].v + points[1].ov) - ((points[1].u + points[1].ou) - (points[0].u + points[0].ou)) - (points[1].v - (points[1].u - points[0].u))
			});
		}
	}

	vox = new Matrix((int)points.size() + 3, 1);
	voy = new Matrix((int)points.size() + 3, 1);
	L = new Matrix((int)points.size() + 3, (int)points.size() + 3);
	vx = new Matrix((int)points.size() + 3, 1);
	vy = new Matrix((int)points.size() + 3, 1);

	int i, j;
	double t, d;

	if (invert) {
		for (auto pt = points.begin(); pt < points.end(); ++pt) {
			*pt = {
				pt->u + pt->ou,
				pt->v + pt->ov,
				-pt->ou,
				-pt->ov
			};
		}
	}

	for (i = 0; i < (int)points.size(); i++) {
		for (j = 0; j < (int)points.size(); j++) {
			t = (points[i].u - points[j].u);
			d = t * t;
			t = (points[i].v - points[j].v);
			d += t * t;
			if (d > 0) d = d * log(d);
			L->table[i][j] = d;
		}

		L->table[i][(int)points.size()] = 1;
		L->table[i][(int)points.size() + 1] = points[i].u;
		L->table[i][(int)points.size() + 2] = points[i].v;

		L->table[(int)points.size()][i] = 1;
		L->table[(int)points.size() + 1][i] = points[i].u;
		L->table[(int)points.size() + 2][i] = points[i].v;

		vx->table[i][0] = points[i].ou;
		vy->table[i][0] = points[i].ov;
	}

//	L->Debug();

	try {
		L->Invert();
	} catch (const char* e) {
		env->ThrowError("ColourWarp: %s", e);
	}

	Matrix::Multiply(L, vx, vox);
	Matrix::Multiply(L, vy, voy);

//	vx->Debug();
//	vy->Debug();
//	L->Debug();
//	vox->Debug();
//	voy->Debug();

	double ox, oy;

	if (!debug) {
		uv_map = (byte_pair*)malloc(sizeof(byte_pair) << 16);
		byte_pair* uv_map_p = uv_map;

		for (int y = 0; y < 256; ++y) {
			for (int x = 0; x < 256; ++x) {
				Warp(x, y, ox, oy);

				if (ox < 0) ox = 0; else if (ox > 255) ox = 255;
				if (oy < 0) oy = 0; else if (oy > 255) oy = 255;
				*uv_map_p++ = { (byte)(ox + 0.5), (byte)(oy + 0.5) };
			}
		}
	} else {
		vi.fps_numerator = 30;
		vi.fps_denominator = 1;
		vi.width = 768;
		vi.height = 768;
		vi.nchannels = 0;
		vi.pixel_type = VideoInfo::CS_YV24;
		vi.num_frames = 1;
		debug_frame = env->NewVideoFrame(vi);

		byte* y_ptr = debug_frame->GetWritePtr(PLANAR_Y);
		byte* u_ptr = debug_frame->GetWritePtr(PLANAR_U);
		byte* v_ptr = debug_frame->GetWritePtr(PLANAR_V);
		int y_pitch = debug_frame->GetPitch(PLANAR_Y);
		int u_pitch = debug_frame->GetPitch(PLANAR_U);
		int v_pitch = debug_frame->GetPitch(PLANAR_V);

		for (int y = 0; y < 768; ++y) {
			if (y == 256 || y == 511) {
				for (int x = 0; x < 768; ++x) {
					y_ptr[x] = 255;
					u_ptr[x] = 192;
					v_ptr[x] = 0;
				}
			} else {
				for (int x = 0; x < 768; ++x) {
					if (x == 256 || x == 511) {
						y_ptr[x] = 255;
						u_ptr[x] = 192;
						v_ptr[x] = 0;
					} else {
						Warp(x - 256, 512 - y, ox, oy);
						ox += 0.5; oy += 0.5;
						if (ox < 0 || ox >= 256 || oy < 0 || oy >= 256) {
							y_ptr[x] = 0;
							u_ptr[x] = v_ptr[x] = 0x80;
						} else {
//							128
//								x 128 - x 128 < ? : 0.5 * : 1.0 128 y 128 - abs - 512 / +*. **
//								y 128 - **
//								+
//								sqrt 181 /
//								128 *
//								-

							if ((!(((int)ox)&0xf) && (((int)ox)&0x10)) || (!(((int)oy)&0xf) && (((int)oy)&0x10))) {
								y_ptr[x] = 255;
								u_ptr[x] = v_ptr[x] = 0x80;
							} else {
								double xx = (ox - 128.5)*(ox < 128.5 ? 0.5 : 1 + (0 - abs(oy) / 512));
								double yy = oy - 128.5;
								int shift = (x < 256 || x >= 512 || y < 256 || y >= 512) ? 1 : 0;
								y_ptr[x] = (byte)(128 - sqrt(xx * xx + yy * yy) / 181 * 128 + 0.5) >> shift;
								u_ptr[x] = ((byte)(ox) >> shift) + (shift ? 64 : 0);
								v_ptr[x] = ((byte)(oy) >> shift) + (shift ? 64 : 0);
							}
						}
					}
				}
			}

			y_ptr += y_pitch;
			u_ptr += u_pitch;
			v_ptr += v_pitch;
		}
	}

	delete vox, voy, L, vx, vy;
}

ColourWarp::~ColourWarp() {
	free(uv_map);
	points.clear();
}

PVideoFrame __stdcall ColourWarp::GetFrame(int n, IScriptEnvironment* env) {
	if (debug) return debug_frame;

	PVideoFrame src = child->GetFrame(n, env);
	PVideoFrame dst = env->NewVideoFrame(vi);

	const byte* src_y = src->GetReadPtr(PLANAR_Y);
	byte* dst_y = dst->GetWritePtr(PLANAR_Y);
	int src_pitch = src->GetPitch(PLANAR_Y);
	int dst_pitch = dst->GetPitch(PLANAR_Y);

	if (src_pitch == dst_pitch) {
		memcpy(dst_y, src_y, src_pitch*vi.height);
	} else {
		for (int y = 0; y < vi.height; ++y) {
			memcpy(dst_y, src_y, vi.width);
			dst_y += dst_pitch;
			src_y += src_pitch;
		}
	}

	const byte* src_u = src->GetReadPtr(PLANAR_U);
	const byte* src_v = src->GetReadPtr(PLANAR_V);
	byte* dst_u = dst->GetWritePtr(PLANAR_U);
	byte* dst_v = dst->GetWritePtr(PLANAR_V);
	int src_pitch_u = src->GetPitch(PLANAR_U);
	int src_pitch_v = src->GetPitch(PLANAR_V);
	int dst_pitch_u = dst->GetPitch(PLANAR_U);
	int dst_pitch_v = dst->GetPitch(PLANAR_V);

	int uv_width = vi.width >> vi.GetPlaneWidthSubsampling(PLANAR_U);
	int uv_height = vi.height >> vi.GetPlaneWidthSubsampling(PLANAR_U);

	for (int y = 0; y < uv_height; ++y) {
		for (int x = 0; x < uv_width; ++x) {
//			int offset = (src_u[x] << 8) | src_v[x];
//			debug("%d", offset);
			byte_pair* p = &uv_map[(src_v[x] << 8) | src_u[x]];
			dst_u[x] = p->u;
			dst_v[x] = p->v;
		}
		src_u += src_pitch_u;
		src_v += src_pitch_v;
		dst_u += dst_pitch_u;
		dst_v += dst_pitch_v;
	}

	return dst;
}

AVSValue __cdecl Create_ColourWarp(AVSValue args, void* user_data, IScriptEnvironment* env) {
	return new ColourWarp(
		args[0].AsClip(),
		args[1],
		args[2],
		args[3],
		args[4].AsBool(false),
		args[5].AsBool(false),
		env
	);
}

#ifndef AVS_STUDIO
const AVS_Linkage *AVS_linkage = 0;
#endif

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit3(IScriptEnvironment* env, const AVS_Linkage* const vectors) {
#ifndef AVS_STUDIO
	AVS_linkage = vectors;
#endif
	env->AddFunction("ColorWarp" , "ci*c*i*[invert]b[debug]b", Create_ColourWarp, 0);
	env->AddFunction("ColourWarp", "ci*c*i*[invert]b[debug]b", Create_ColourWarp, 0);

	return "`ColourWarp' ColourWarp plugin";
}
